/*
 *      SCCS:  @(#)getsys.c	1.11 (96/11/04) 
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1992 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 *
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)getsys.c	1.11 (96/11/04) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)getsys.c	1.11 96/11/04 TETware release 3.2
NAME:		getsys.c
PRODUCT:	TETware
AUTHOR:		Andrew Dingwall, UniSoft Ltd.
DATE CREATED:	April 1992

SYNOPSIS:
	#include "tet_api.h"
	int tet_remgetsys(void);
	int tet_getsysbyid(int sysid, struct tet_sysent *sysp);

DESCRIPTION:
	DTET API functions

	Tet_remgetsys() returns the (numeric) system name of the local
	system.  The TETware-Lite version always returns zero.

	Tet_getsysbyid() places the `systems' file entry for the
	specified system ID into the space pointed to by sysp.
	It is not supported in TETware-Lite.

MODIFICATIONS:
	Andrew Dingwall, UniSoft Ltd., December 1993
	changed dapi.h to dtet2/tet_api.h

	Geoff Clare, UniSoft Ltd., July 1996
	Changes for TETWare.

	Geoff Clare, UniSoft Ltd., Sept 1996
	Changes for TETWare-Lite.

************************************************************************/

#ifndef TET_LITE /* -START-LITE-CUT- */

#include <string.h>
#include "dtmac.h"
#include "dtthr.h"
#include "sysent.h"
#include "tet_api.h"

#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif

int tet_remgetsys()
{
	extern int tet_mysysid;

	return(tet_mysysid);
}

int tet_getsysbyid(sysid, sysp)
int sysid;
struct tet_sysent *sysp;
{
	struct sysent *entp;

	if (!sysp)
	{
		tet_errno = TET_ER_INVAL;
		return -1;
	}

	API_LOCK;

	entp = tet_libgetsysbyid(sysid);
	if (!entp)
	{
		tet_errno = TET_ER_SYSID;
		API_UNLOCK;
		return -1;
	}

	/* copy the sy_sysid */
	sysp->ts_sysid = entp->sy_sysid;

	/* copy the sy_name - the sizes must be the same */
#if TET_SNAMELEN != SNAMELEN
#error "TET_SNAMELEN != SNAMELEN"
#endif
	(void) strncpy(sysp->ts_name, entp->sy_name, sizeof(sysp->ts_name));
	sysp->ts_name[sizeof(sysp->ts_name) - 1] = '\0';

	/* ignore the sy_tccd */
	
	API_UNLOCK;
	return 0;
}

#else /* -END-LITE-CUT- */

int tet_remgetsys()
{
	return 0;
}

#endif /* -LITE-CUT-LINE- */
