/*
 *	SCCS: @(#)mutexes.c	1.3 (96/09/30)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)mutexes.c	1.3 (96/09/30) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)mutexes.c	1.3 96/09/30 TETware release 3.2
NAME:		mutex definitions and handling functions
PRODUCT:	TETware
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	July 1996
SYNOPSIS:

	mutex_t	tet_top_mtx;
	mutex_t tet_thrtab_mtx;
	mutex_t tet_thrwait_mtx;
	mutex_t tet_alarm_mtx;
	mutex_t tet_sigalrm_mtx;

	void	tet_mtx_init(void);
	void	tet_mtx_destroy(void);
	void	tet_mtx_lock(void);
	void	tet_mtx_unlock(void);

DESCRIPTION:

	Tet_mtx_init() and tet_mtx_destroy() are used by the TCM to
	initialise/destroy/reinitialise all mutexes between test
	purposes and before/after the startup and cleanup functions.

	Tet_mtx_lock() and tet_mtx_unlock() are used to lock/unlock
	all mutexes (e.g. before a fork1()).  The locking is done in
	a specific order.  If any code needs to have more than one
	mutex locked at the same time it must lock them in the
	same order as tet_mtx_lock().

MODIFICATIONS:

************************************************************************/

#include "dtmac.h"
#include "dtthr.h"

mutex_t tet_top_mtx;		/* used in tet_api_lock() */
mutex_t tet_thrtab_mtx;		/* mutex for thread table */
mutex_t tet_thrwait_mtx;	/* used in tet_cln_threads() */
mutex_t tet_sigalrm_mtx;	/* used in tet_set_alarm() etc. */
mutex_t tet_alarm_mtx;		/* used in tet_set_alarm() etc. */

void
tet_mtx_init()
{
	/* initialise all mutexes */
	(void) mutex_init(&tet_top_mtx, USYNC_THREAD, (void *)0);
	(void) mutex_init(&tet_thrtab_mtx, USYNC_THREAD, (void *)0);
	(void) mutex_init(&tet_thrwait_mtx, USYNC_THREAD, (void *)0);
	(void) mutex_init(&tet_sigalrm_mtx, USYNC_THREAD, (void *)0);
	(void) mutex_init(&tet_alarm_mtx, USYNC_THREAD, (void *)0);
}

void
tet_mtx_destroy()
{
	/* destroy all mutexes */

	(void) mutex_destroy(&tet_top_mtx);
	(void) mutex_destroy(&tet_thrtab_mtx);
	(void) mutex_destroy(&tet_thrwait_mtx);
	(void) mutex_destroy(&tet_sigalrm_mtx);
	(void) mutex_destroy(&tet_alarm_mtx);
}

void
tet_mtx_lock()
{
	/* lock all mutexes IN THE CORRECT ORDER */

	(void) mutex_lock(&tet_top_mtx);	/* must be first */

	(void) mutex_lock(&tet_thrtab_mtx);
	(void) mutex_lock(&tet_thrwait_mtx);	/* after thrtab_mtx */

	(void) mutex_lock(&tet_sigalrm_mtx);
	(void) mutex_lock(&tet_alarm_mtx);	/* after sigalrm_mtx */
}

void
tet_mtx_unlock()
{
	/* unlock all mutexes */

	(void) mutex_unlock(&tet_alarm_mtx);
	(void) mutex_unlock(&tet_sigalrm_mtx);

	(void) mutex_unlock(&tet_thrwait_mtx);
	(void) mutex_unlock(&tet_thrtab_mtx);

	(void) mutex_unlock(&tet_top_mtx);
}
