/*
 *	SCCS: @(#)dtthr.h	1.15 (97/07/21)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

/************************************************************************

SCCS:   	@(#)dtthr.h	1.15 97/07/21 TETware release 3.2
NAME:		dtthr.h
PRODUCT:	TETware
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	July 1996

DESCRIPTION:
	macros used for thread safety

MODIFICATIONS:
	
	Geoff Clare, UniSoft Ltd., Sept 1996
	Moved alarm stuff to alarm.h.
	Changes for POSIX threads.
	
	Geoff Clare, UniSoft Ltd., Oct 1996
	Added tet_thr_equal definitions.
	
	Geoff Clare, UniSoft Ltd., June 1997
	Changes to support NT threads.

************************************************************************/

#ifdef TET_THREADS


#    include <signal.h>

#    ifndef TET_POSIX_THREADS

#      include <thread.h>
#      include <synch.h>

#      define tet_thr_equal(a,b)	((a) == (b))

#    else /* TET_POSIX_THREADS */

#      include <sys/types.h>
#      include <pthread.h>

#      define tet_thr_equal(a,b)	pthread_equal(a, b)

       /* Define the UI thread subset that we use in terms of pthreads */

#define cond_t pthread_cond_t
#define thread_key_t pthread_key_t
#define mutex_t pthread_mutex_t
#define thread_t pthread_t
#define timestruc_t  struct timespec

#      define	USYNC_THREAD		0

       /* we only use USYNC_THREAD cond vars, so default POSIX attr is OK */
#      define cond_init(cvp,t,a) \
			pthread_cond_init((cvp),(pthread_condattr_t *)0)
#      define cond_destroy	pthread_cond_destroy
#      define cond_signal	pthread_cond_signal
#      define cond_timedwait	pthread_cond_timedwait

       /* we only use USYNC_THREAD mutexes, so default POSIX attr is OK */
#      define mutex_init(mp,t,a) \
			pthread_mutex_init((mp), (pthread_mutexattr_t *)0)
#      define mutex_destroy	pthread_mutex_destroy
#      define mutex_lock	pthread_mutex_lock
#      define mutex_unlock	pthread_mutex_unlock

       /* we only use default thread attrs (except in tet_pthread_create) */
#      define thr_create(sp,ss,func,arg,fl,tp) \
			pthread_create((tp), (pthread_attr_t *)0, (func), (arg))
#      define thr_exit		pthread_exit

       /* thr_join tid must refer to a specific thread */
#      define thr_join(tid,dtp,sp) \
			pthread_join((dtp)?(*(dtp)=(tid)):(tid), (sp))
#      define thr_keycreate	pthread_key_create
#      define thr_kill		pthread_kill
#      define thr_self		pthread_self
#      define thr_getspecific(key,vp) \
			((*(vp) = pthread_getspecific(key)), 0)
#      define thr_setspecific	pthread_setspecific
#      define thr_sigsetmask	pthread_sigmask

#    endif /* TET_POSIX_THREADS */


   /* Keys for thread-specific data */
   extern thread_key_t tet_errno_key;
   extern thread_key_t tet_block_key;
   extern thread_key_t tet_sequence_key;
     extern thread_key_t tet_child_key;
     extern thread_key_t tet_alrm_flag_key;


     /* set of all blockable signals (to be blocked during critical code) */
     extern sigset_t tet_blockable_sigs;

     /* macros for signal-safe mutex locking */
     /* These must be used in matching pairs as if they were braces */
#    define MTX_LOCK(mp) { sigset_t MTX_LOCK_oss; int MTX_LOCK_maskret = \
	thr_sigsetmask(SIG_BLOCK, &tet_blockable_sigs, &MTX_LOCK_oss); \
	(void) mutex_lock(mp);

#    define MTX_UNLOCK(mp) (void) mutex_unlock(mp); if (MTX_LOCK_maskret == 0) \
	(void) thr_sigsetmask(SIG_SETMASK, &MTX_LOCK_oss, (sigset_t *)0); }


   /* top-level API mutex with calls that can be nested */
#  define API_LOCK	tet_api_lock(1, srcFile, __LINE__)
#  define API_UNLOCK	tet_api_lock(0, srcFile, __LINE__)
   extern void tet_api_lock();

#  ifndef NEEDsrcFile
#    define NEEDsrcFile
#  endif

#else /* !TET_THREADS */

  /* allow use of symbols without putting them in #ifdef TET_THREADS */

#  define MTX_LOCK(mp)		{
#  define MTX_UNLOCK(mp)	}
#  define API_LOCK
#  define API_UNLOCK

   extern long tet_sequence;

#endif /* !TET_THREADS */

