/*
 *	SCCS: @(#)config.c	1.9 (97/08/12)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)config.c	1.9 (97/08/12) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)config.c	1.9 97/08/12 TETware release 3.2
NAME:		config.c
PRODUCT:	TETware
AUTHOR:		Andrew Dingwall, UniSoft Ltd.
DATE CREATED:	August 1996

DESCRIPTION:
	functions which deal with configuration variables

MODIFICATIONS:
	Geoff Clare, UniSoft Ltd., August 1996
	Missing <string.h>.

	Andrew Dingwall, UniSoft Ltd., June 1997.
	added get_runtime_tsroot() to return the runtime value of
	TET_TSROOT - required to support the correct operation of TET_RUN
	when a remote system's test suite root directory resides on
	a read-only file system

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#  include <unistd.h>
#include "dtmac.h"
#include "dtmsg.h"
#include "ptab.h"
#include "error.h"
#include "ltoa.h"
#include "config.h"
#include "servlib.h"
#include "dtetlib.h"
#include "systab.h"
#include "tcc.h"
#include "dtcc.h"

#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif

/* default config file names on each system */
static char *defcfname[] = {
	"tetbuild.cfg",
	"tetexec.cfg",
	"tetclean.cfg",
#ifndef TET_LITE	/* -START-LITE-CUT- */
	"tetdist.cfg"
#endif /* !TET_LITE */	/* -END-LITE-CUT- */
};

/* the default config file name for a particular mode */
#define DEFCFNAME(mode)	(defcfname[TC_CONF_MODE(mode)])

/* config file names - these get filled in by initcfg() */
static char *cfname[sizeof defcfname / sizeof defcfname[0]];

/* the actual config file name for a particular mode */
#define CFNAME(mode)	(cfname[TC_CONF_MODE(mode)])

/*
** the master configuration lists
**
** there are master lists for: build mode, execute mode, clean mode,
** the so-called distributed variables and variables specified with -v
** on the command-line
**
** for each list there is a pointer, a length and the number of entries in
** the list
**
** the per-system lists are maintained in the systab entry for each system
*/
static struct cflist mcflist[TC_NCONF_MODES + 1];
static struct cflist vopts;

/* the master config list for a particular mode */
#define MCFLIST(mode)	(mcflist[TC_CONF_MODE(mode)])

/* config error counter */
static int conferrors;


#ifdef TET_LITE /* -LITE-CUT-LINE- */

/*
** file names for the temporary master config files which are passed
** to test cases via the TET_CONFIG communication variable
*/
static char *tcfname[sizeof defcfname / sizeof defcfname[0]];
#define Ntcfname	(sizeof tcfname / sizeof tcfname[0])
#define TCFNAME(mode)	(tcfname[TC_CONF_MODE(mode)])

#else /* -START-LITE-CUT- */

/*
** file names for the temporary master exec, distrib and command-line
** config files which are passed to XRESD in order to support tet_remexec()
*/
static char *ecfname, *dcfname, *ccfname;

#endif /* TET_LITE */	/* -END-LITE-CUT- */


/* static function declarations */
static void addvopts PROTOLIST((struct cflist *));
static int cfix2 PROTOLIST((int));
static int cflag2bool PROTOLIST((char *, char *));
static void checkbvar PROTOLIST((struct cflist *, int, char *));
static void compat_fix PROTOLIST((void));
static void confgiveup PROTOLIST((void));
static int docff2 PROTOLIST((char *, struct cflist *, char *));
static char *docffile PROTOLIST((struct cflist *, char *));
static void docfl2 PROTOLIST((struct cflist *, struct cflist *, char *));
static void docfloc PROTOLIST((struct systab *));
static char **findcfg PROTOLIST((char *, struct cflist *));
static void fix_tet_api_compliant PROTOLIST((int));
static void fix_tet_pass_tc_name PROTOLIST((int));
static void proccfl2 PROTOLIST((char *, struct cflist *, int));
static void proccfline PROTOLIST((char *, struct cflist *, int, char *));
static void readmconf PROTOLIST((char *, struct cflist *));
static void reportcfg PROTOLIST((struct cflist *, int, int));
static void rptconf PROTOLIST((struct cflist *));
#ifndef TET_LITE	/* -START-LITE-CUT- */
static void docfr2 PROTOLIST((struct cflist *, struct cflist *,
	struct systab *, int, char *));
static void docfrem PROTOLIST((struct systab *));
static void rptdist PROTOLIST((struct cflist *));
#endif /* !TET_LITE */	/* -END-LITE-CUT- */


/*
**	initcfg() - perform initial configuration
**
**	fopt is the -f command-line option
**	gopt is the -g command-line option
**	xopt is the -x command-line option
**	cwd is tcc's initial working directory
*/

void initcfg(fopt, gopt, xopt, cwd)
char *fopt, *gopt, *xopt, *cwd;
{
	static char tet_version[] =
#ifdef TET_LITE	/* -LITE-CUT-LINE- */
					"TET_VERSION=3.2-lite";
#else		/* -START-LITE-CUT- */
					"TET_VERSION=3.2";
#endif		/* -END-LITE-CUT- */
	char fname[MAXPATH];

	TRACE5(tet_Ttcc, 2, "initcfg(): fopt = \"%s\", gopt = \"%s\", xopt = \"%s\", cwd = \"%s\"",
		fopt ? fopt : "", gopt ? gopt : "", xopt ? xopt : "", cwd);

	/* read in the configuration variables for each of the chosen
		modes of operation */
	if (tcc_modes & TCC_BUILD) {
		if (gopt && *gopt)
			fullpath(cwd, gopt, fname, sizeof fname, 0);
		else
			fullpath(tet_tsroot, DEFCFNAME(CONF_BUILD), fname,
				sizeof fname, 0);
		readmconf(fname, &MCFLIST(CONF_BUILD));
		CFNAME(CONF_BUILD) = rstrstore(fname);
		TRACE2(tet_Ttcc, 1, "build config file = %s",
			CFNAME(CONF_BUILD));
	}
	if (tcc_modes & TCC_EXEC) {
		if (xopt && *xopt)
			fullpath(cwd, xopt, fname, sizeof fname, 0);
		else if (tet_execute && *tet_execute) {
			fullpath(tet_execute, DEFCFNAME(CONF_EXEC), fname,
				sizeof fname, 0);
			if (tet_eaccess(fname, 0) < 0)
				fullpath(tet_tsroot, DEFCFNAME(CONF_EXEC),
					fname, sizeof fname, 0);
		}
		else
			fullpath(tet_tsroot, DEFCFNAME(CONF_EXEC), fname,
				sizeof fname, 0);
		readmconf(fname, &MCFLIST(CONF_EXEC));
		CFNAME(CONF_EXEC) = rstrstore(fname);
		TRACE2(tet_Ttcc, 1, "exec config file = %s",
			CFNAME(CONF_EXEC));
	}
	if (tcc_modes & TCC_CLEAN) {
		if (fopt && *fopt)
			fullpath(cwd, fopt, fname, sizeof fname, 0);
		else
			fullpath(tet_tsroot, DEFCFNAME(CONF_CLEAN), fname,
				sizeof fname, 0);
		readmconf(fname, &MCFLIST(CONF_CLEAN));
		CFNAME(CONF_CLEAN) = rstrstore(fname);
		TRACE2(tet_Ttcc, 1, "clean config file = %s",
			CFNAME(CONF_CLEAN));
	}

	if (conferrors)
		confgiveup();

	/*
	** add in the command-line variables
	** supply default TET_API_COMPLIANT and TET_PASS_TC_NAME
	** add in TET_VERSION
	*/
	if (tcc_modes & TCC_BUILD) {
		addvopts(&MCFLIST(CONF_BUILD));
		fix_tet_api_compliant(CONF_BUILD);
		fix_tet_pass_tc_name(CONF_BUILD);
		proccfl2(tet_version, &MCFLIST(CONF_BUILD), 1);
	}
	if (tcc_modes & TCC_EXEC) {
		addvopts(&MCFLIST(CONF_EXEC));
		fix_tet_api_compliant(CONF_EXEC);
		fix_tet_pass_tc_name(CONF_EXEC);
		proccfl2(tet_version, &MCFLIST(CONF_EXEC), 1);
	}
	if (tcc_modes & TCC_CLEAN) {
		addvopts(&MCFLIST(CONF_CLEAN));
		fix_tet_api_compliant(CONF_CLEAN);
		fix_tet_pass_tc_name(CONF_CLEAN);
		proccfl2(tet_version, &MCFLIST(CONF_CLEAN), 1);
	}

	/* fix up tet_compat */
	compat_fix();

	if (conferrors)
		confgiveup();

	TRACE1(tet_Ttcc, 2, "initcfg() RETURN");
}

/*
**	readmconf() - read in a set of master configuration variables
**		from a configuration file
*/

static void readmconf(fname, lp)
char *fname;
struct cflist *lp;
{
	FILE *fp;
	char buf[MAXPATH * 2];
	register char *p;
	int lcount = 0;

	/* open the config file */
	if ((fp = fopen(fname, "r")) == (FILE *) 0) {
		error(errno, "can't open", fname);
		conferrors++;
		return;
	}

	/* process each line in turn -
	** ignore blank lines and comments
	** trim trailing spaces
	*/
	while (fgets(buf, sizeof buf, fp) != (char *) 0) {
		lcount++;
		for (p = buf; *p; p++)
			if (*p == '\n' || *p == '#') {
				*p = '\0';
				break;
			}
		while (--p >= buf)
			if (isspace(*p))
				*p = '\0';
			else
				break;
		if (p >= buf)
			proccfline(buf, lp, lcount, fname);
	}

	(void) fclose(fp);
}

/*
**	proccfline() - process a single configuration line
*/

static void proccfline(line, lp, lineno, fname)
char *line, *fname;
struct cflist *lp;
int lineno;
{
	static char fmt[] =
		"bad format config variable assignment at line %d in file";
	char msg[sizeof fmt + LNUMSZ];

	/* check the format of the config line */
	if (!tet_equindex(line) || !tet_remvar(line, -1)) {
		(void) sprintf(msg, fmt, lineno);
		error(0, msg, fname);
		conferrors++;
		return;
	}

	/* perform common variable assignment processing */
	proccfl2(line, lp, 1);
}

/*
**	proccfl2() - common config line processing
*/

static void proccfl2(line, lp, store)
char *line;
register struct cflist *lp;
int store;
{
	register char **cp, *p;

	/* add a new value or update an existing one */
	if ((cp = findcfg(line, lp)) == (char **) 0) {
		RBUFCHK((char **) &lp->cf_conf, &lp->cf_lconf,
			(int) ((lp->cf_nconf + 1) * sizeof *lp->cf_conf));
		p = store ? rstrstore(line) : line;
		*(lp->cf_conf + lp->cf_nconf++) = p;
	}
	else {
		ASSERT(*cp);
		if (strcmp(line, *cp)) {
			if (store) {
				TRACE2(tet_Tbuf, 6, "free config line = %s",
					tet_i2x(*cp));
				free(*cp);
				*cp = rstrstore(line);
			}
			else
				*cp = line;
		}
	}
}

/*
**	findcfg() - find a config list entry in a configuration list
**		for the specified name and return a pointer to the
**		entry's address
**
**	name may be delimited either by '=' or '\0'
**
**	return (char **) 0 if no match can be found
*/

static char **findcfg(name, lp)
char *name;
register struct cflist *lp;
{
	register int len;
	register char **cp, *p;

	/* make len equal to the number of bytes in the name */
	if ((p = tet_equindex(name)) == (char *) 0)
		len = strlen(name);
	else
		len = p - name;

	/*
	** search the list for an entry whose initial part matches the
	** specified name
	*/
	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
		if (*cp && !strncmp(*cp, name, len) && *(*cp + len) == '=')
			return(cp);

	return((char **) 0);
}

/*
**	addvopts() - add the command-line variables to a configuration
**		variable list
*/

static void addvopts(lp)
struct cflist *lp;
{
	register char **vp;

	TRACE1(tet_Ttcc, 6, "addvopts()");

	for (vp = vopts.cf_conf; vp < vopts.cf_conf + vopts.cf_nconf; vp++)
		proccfl2(*vp, lp, 1);

	TRACE1(tet_Ttcc, 6, "addvopts() RETURN");
}

/*
**	fix_tet_api_compliant() - provide a default value for TET_API_COMPLIANT
*/

static void fix_tet_api_compliant(mode)
int mode;
{
	static char name[] = "TET_API_COMPLIANT";
	char line[sizeof name + 6];

	TRACE2(tet_Ttcc, 6, "fix_tet_api_compliant(%s)", prcfmode(mode));

	ASSERT(CONF_MODE_OK(mode, mcflist));

	/* return now if TET_API_COMPLIANT is already defined */
	if (getmcfg(name, mode))
		return;

	/* here to add a default value of !TET_OUTPUT_CAPTURE */
	(void) sprintf(line, "%s=%s", name,
		getmcflag("TET_OUTPUT_CAPTURE", mode) ? "False" : "True");
	proccfl2(line, &MCFLIST(mode), 1);
}

/*
**	fix_tet_pass_tc_name() - provide a default value for TET_PASS_TC_NAME
*/

static void fix_tet_pass_tc_name(mode)
int mode;
{
	static char name[] = "TET_PASS_TC_NAME";
	char line[sizeof name + 6];

	TRACE2(tet_Ttcc, 6, "fix_pass_tc_name(%s)", prcfmode(mode));

	ASSERT(CONF_MODE_OK(mode, mcflist));

	/* return now if TET_PASS_TC_NAME is already defined */
	if (getmcfg(name, mode))
		return;

	/* here to add a default value of TET_OUTPUT_CAPTURE */
	(void) sprintf(line, "%s=%s", name,
		getmcflag("TET_OUTPUT_CAPTURE", mode) ? "True" : "False");
	proccfl2(line, &MCFLIST(mode), 1);
}

/*
**	compat_fix() - fix up the value of tet_compat
**
**	the value of TET_COMPAT is extracted from the master config list
**	for each of the selected modes of operation
*/

static void compat_fix()
{
	int compat;
	int conflict;

	TRACE1(tet_Ttcc, 6, "compat_fix()");

	/* get the value of TET_COMPAT from each of the master
	** configurations and check for conflicts
	*/
	tet_compat = cfix2(CONF_BUILD);
	conflict = 0;

	if ((compat = cfix2(CONF_EXEC)) != 0) {
		if (tet_compat && compat != tet_compat)
			conflict++;
		tet_compat = compat;
	}

	if ((compat = cfix2(CONF_CLEAN)) != 0) {
		if (tet_compat && compat != tet_compat)
			conflict++;
		tet_compat = compat;
	}

	/* report a conflict if one has been detected */
	if (conflict) {
		error(0, "conflicting values for TET_COMPAT",
			"have been specified");
		conferrors++;
	}

#ifndef NOTRACE
	if (tet_compat)
		TRACE2(tet_Tscen, 1, "running in %sTET compatibility mode",
			tet_compat == COMPAT_ETET ? "E" : "D");
	else
		TRACE1(tet_Tscen, 1, "no compatibility mode specified");
#endif
}

/*
**	cfix2() - extend the compat_fix() processing for a particular
**		master configuration list
**
**	return COMPAT_DTET or COMPAT_ETET if TET_COMPAT has been so
**	defined, or 0 if TET_COMPAT has not been defined with a valid value
*/

static int cfix2(mode)
int mode;
{
	static char compatname[] = "TET_COMPAT";
	register char *p;

	if ((p = getmcfg(compatname, mode)) != (char *) 0)
		switch (*p) {
		case 'D':
		case 'd':
			return(COMPAT_DTET);
		case 'E':
		case 'e':
			return(COMPAT_ETET);
		default:
			error(0, compatname, "variable has ambiguous value");
			conferrors++;
			break;
		}

	return(0);
}

/*
**	distcfg() - process the distributed configuration
*/

void distcfg()
{
	/*
	** environment variables on the master system which are added
	** to the distributed configuration
	**
	** note that the number and order of these variables must
	** correspond to the initialisation code below
	*/
	static struct dvar {
		char *dv_name;		/* variable name */
		int dv_needed;		/* variable is needed on each system */
		int dv_len;		/* strlen(mv_name) */
		char *dv_value;		/* value on master system */
	} dvar[] = {
		{ "TET_ROOT",		1 },
		{ "TET_SUITE_ROOT",	0 },
		{ "TET_TSROOT",		1 },
		{ "TET_EXECUTE",	0 },
		{ "TET_TMP_DIR",	0 },
		{ "TET_RUN",		0 }
	};

#define Ndvar	(sizeof dvar / sizeof dvar[0])

	static char tet_rem[] = "TET_REM000_";

	register struct dvar *dvp;
	register int sysid, sysmax;
	register char *p1;
	char buf[MAXPATH + 40];
	static char fmt1[] =
		"%s distributed configuration variable for system %d";
	static char fmt2[] =
		"%s distributed configuration variable not defined for system";
#ifndef TET_LITE	/* -START-LITE-CUT- */
	register char *p2, **cp;
	char fname[MAXPATH];
#endif /* !TET_LITE */	/* -END-LITE-CUT- */

	TRACE1(tet_Ttcc, 2, "distcfg(): process distributed configuration");

	/*
	** set up dvar first time through -
	** the number and order of these assignments must
	** correspond to the variables in the dvar array above
	*/
	if ((dvp = dvar)->dv_value == (char *) 0) {
		(dvp++)->dv_value = tet_root;
		(dvp++)->dv_value = tet_suite_root;
		(dvp++)->dv_value = tet_tsroot;
		(dvp++)->dv_value = tet_execute;
		(dvp++)->dv_value = tet_tmp_dir;
		(dvp++)->dv_value = tet_run;
		for (dvp = dvar; dvp < dvar + Ndvar; dvp++)
			dvp->dv_len = strlen(dvp->dv_name);
	}

	/* add in the (non-empty) dist variables for the master system */
	for (dvp = dvar; dvp < dvar + Ndvar; dvp++) {
		if (!dvp->dv_value || !*dvp->dv_value)
			continue;
		(void) sprintf(buf, "%s%s=%.*s", tet_rem, dvp->dv_name,
			(int) sizeof buf - (int) sizeof tet_rem - (int) strlen(dvp->dv_name) - 1,
			dvp->dv_value);
		proccfl2(buf, &MCFLIST(CONF_DIST), 1);
	}


#ifndef TET_LITE	/* -START-LITE-CUT- */

	/*
	** if we need the distributed config file, read it in; then add in
	** distributed variables with nnn > 0 from the tcc command line set
	*/
	if (ts_needdist() || symax() > 0) {
		fullpath(tet_tsroot, DEFCFNAME(CONF_DIST), fname,
			sizeof fname, 0);
		readmconf(fname, &MCFLIST(CONF_DIST));
		CFNAME(CONF_DIST) = rstrstore(fname);
		TRACE2(tet_Ttcc, 1, "distributed config file = %s",
			CFNAME(CONF_DIST));
		for (cp = vopts.cf_conf; cp < vopts.cf_conf + vopts.cf_nconf; cp++) {
			if ((p1 = tet_remvar(*cp, -1)) == *cp ||
				tet_remvar(*cp, 0) != *cp)
					continue;
			ASSERT(p1);
			p2 = tet_equindex(p1);
			ASSERT(p2);
			for (dvp = dvar; dvp < dvar + Ndvar; dvp++)
				if (
					dvp->dv_len == p2 - p1 &&
					!strncmp(p1, dvp->dv_name, dvp->dv_len)
				)
					proccfl2(*cp, &MCFLIST(CONF_DIST), 1);
		}
	}

#endif /* !TET_LITE */	/* -END-LITE-CUT- */


	/*
	** ensure that we have all the dist config variables that we need
	** and that they refer to full path names
	*/
	for (sysid = 0, sysmax = symax(); sysid <= sysmax; sysid++) {
		if (!syfind(sysid))
			continue;
		for (dvp = dvar; dvp < dvar + Ndvar; dvp++) {
			p1 = getdcfg(dvp->dv_name, sysid);
			if (p1 && *p1) {
				if (sysid == 0)
					ASSERT(isabspathloc(p1));
				else if (!isabspathrem(p1)) {
					(void) sprintf(buf, fmt1,
						dvp->dv_name, sysid);
					error(0, buf,
						"is not an absolute path name");
					conferrors++;
				}
			}
			else if (dvp->dv_needed) {
				(void) sprintf(buf, fmt2, dvp->dv_name);
				error(0, buf, tet_i2a(sysid)); 
				conferrors++;
			}
		}
	}

	if (conferrors)
		confgiveup();

	TRACE1(tet_Ttcc, 2, "distcfg() RETURN");
}

/*
**	procvopt() - process a -v command-line option
**
**	return 0 if successful or -1 on error
*/

int procvopt(line)
char *line;
{
	/* check the format of the config line */
	if (!tet_equindex(line) || !tet_remvar(line, -1)) {
		error(0, "-v option is badly formatted:", line);
		return(-1);
	}

	/* perform common variable assignment processing */
	proccfl2(line, &vopts, 1);
	return(0);
}

/*
**	doconfig() - fix up all the configurations and write them to the
**		journal
*/

void doconfig()
{
	register struct systab *sp;
	struct cflist *blp, *elp, *clp;
#ifndef TET_LITE	/* -START-LITE-CUT- */
	register int sysid, sysmax;
#endif /* !TET_LITE */	/* -END-LITE-CUT- */

	TRACE1(tet_Ttcc, 2, "doconfig(): fix up all the configurations and write them to the journal");

	/*
	** if we are processing on the local system, fix up the local configs
	** and arrange to report the local configs to the journal;
	** otherwise, arrange to report the master configs to the journal
	*/
	if ((sp = syfind(0)) != (struct systab *) 0) {
		docfloc(sp);
		blp = &CFLIST(sp, CONF_BUILD);
		elp = &CFLIST(sp, CONF_EXEC);
		clp = &CFLIST(sp, CONF_CLEAN);
	}
	else {
		blp = &MCFLIST(CONF_BUILD);
		elp = &MCFLIST(CONF_EXEC);
		clp = &MCFLIST(CONF_CLEAN);
	}


#ifndef TET_LITE	/* -START-LITE-CUT- */

	/* fix up the remote configs */
	sysmax = symax();
	for (sysid = 1; sysid <= sysmax; sysid++)
		if ((sp = syfind(sysid)) != (struct systab *) 0) 
			docfrem(sp);

#endif /* !TET_LITE */	/* -END-LITE-CUT- */


	/* give up if any of the above failed */
	if (conferrors)
		confgiveup();

	/* report the local/master configs to the journal */
	if (tcc_modes & TCC_BUILD)
		reportcfg(blp, 0, CONF_BUILD);
	if (tcc_modes & TCC_EXEC)
		reportcfg(elp, 0, CONF_EXEC);
	if (tcc_modes & TCC_CLEAN)
		reportcfg(clp, 0, CONF_CLEAN);


#ifdef TET_LITE		/* -LITE-CUT-LINE- */

	/*
	** write out the temporary master config files which are passed
	** to test cases via the TET_CONFIG communication variable
	*/
	if ((tcc_modes & TCC_BUILD) && (TCFNAME(CONF_BUILD) = docffile(&MCFLIST(CONF_BUILD), "master build")) == (char *) 0)
		conferrors++;
	if ((tcc_modes & TCC_EXEC) && (TCFNAME(CONF_EXEC) = docffile(&MCFLIST(CONF_EXEC), "master exec")) == (char *) 0)
		conferrors++;
	if ((tcc_modes & TCC_CLEAN) && (TCFNAME(CONF_CLEAN) = docffile(&MCFLIST(CONF_CLEAN), "master clean")) == (char *) 0)
		conferrors++;

#else			/* -START-LITE-CUT- */

	/* report the dist config */
	if (sysmax > 0 || ts_needdist())
		reportcfg(&MCFLIST(CONF_DIST), 0, CONF_DIST);

	/* report the remote configs */
	for (sysid = 1; sysid <= sysmax; sysid++) {
		if ((sp = syfind(sysid)) == (struct systab *) 0) 
			continue;
		if (tcc_modes & TCC_BUILD)
			reportcfg(&CFLIST(sp, CONF_BUILD), sysid, CONF_BUILD);
		if (tcc_modes & TCC_EXEC)
			reportcfg(&CFLIST(sp, CONF_EXEC), sysid, CONF_EXEC);
		if (tcc_modes & TCC_CLEAN)
			reportcfg(&CFLIST(sp, CONF_CLEAN), sysid, CONF_CLEAN);
	}

	/*
	** write the master exec, distributed and command-line config
	** variables to temporary files for use with tet_remexec()
	*/
	if ((ecfname = docffile(&MCFLIST(CONF_EXEC), "master exec")) == (char *) 0)
		conferrors++;
	if ((dcfname = docffile(&MCFLIST(CONF_DIST), "distributed")) == (char *) 0)
		conferrors++;
	if ((ccfname = docffile(&vopts, "command-line")) == (char *) 0)
		conferrors++;

	/* give up if any of the above failed */
	if (conferrors)
		confgiveup();

	/* then send the file names to XRESD */
	if (tet_xdcfname(ecfname, dcfname, ccfname) < 0) {
		error(tet_xderrno, "can't send config file names to XRESD",
			(char *) 0);
		conferrors++;
	}

#endif /* TET_LITE */	/* -END-LITE-CUT- */


	/* give up if any of the above failed */
	if (conferrors)
		confgiveup();
}

/*
**	docfloc() - fix up the local configs for each of the selected
**		modes of operation
*/

static void docfloc(sp)
register struct systab *sp;
{
	TRACE1(tet_Ttcc, 2, "docfloc(): fix up all the local configurations");

	if (tcc_modes & TCC_BUILD)
		docfl2(&MCFLIST(CONF_BUILD), &CFLIST(sp, CONF_BUILD), "build");

	if (tcc_modes & TCC_EXEC)
		docfl2(&MCFLIST(CONF_EXEC), &CFLIST(sp, CONF_EXEC), "exec");

	if (tcc_modes & TCC_CLEAN)
		docfl2(&MCFLIST(CONF_CLEAN), &CFLIST(sp, CONF_CLEAN), "clean");
}

/*
**	docfl2() - extend the docfloc() processing for a particular mode
*/

static void docfl2(from, to, cftype)
register struct cflist *from, *to;
char *cftype;
{
	register char **cp;
	register char *p;

	TRACE2(tet_Ttcc, 3, "docfl2(): fix up the local %s configuration",
		cftype);

	/*
	** do two passes down the source list;
	**
	** in the first pass we add/update all the non TET_REMnnn_
	** variables into the destination list
	**
	** in the second pass we add/update all the TET_REMnnn_
	** variables for the local system into the destination list without
	** the TET_REMnnn_ prefix
	**
	** note that proccfl2() is called with a zero store argument;
	** this is because the per-system config lists for system 0
	** do not own their variables but instead contain pointers to
	** variables that are owned by the master config lists
	*/

	/* first pass */
	for (cp = from->cf_conf; cp < from->cf_conf + from->cf_nconf; cp++)
		if (tet_remvar(*cp, -1) == *cp)
			proccfl2(*cp, to, 0);

	/* second pass */
	for (cp = from->cf_conf; cp < from->cf_conf + from->cf_nconf; cp++)
		if ((p = tet_remvar(*cp, 0)) != *cp) {
			ASSERT(p);
			proccfl2(p, to, 0);
		}

	/* finally, check all of the boolean variables */
	checkbvar(to, 0, cftype);
}


#ifndef TET_LITE	/* -START-LITE-CUT- */

/*
**	docfrem() - fix up the remote configs for each of the selected
**		modes of operation
*/

static void docfrem(sp)
struct systab *sp;
{
	struct cflist tmp;

	TRACE2(tet_Ttcc, 2,
		"docfrem() - fix up all the configurations on system %s",
		tet_i2a(sp->sy_sysid));

	ASSERT(sp->sy_sysid > 0);

	/* initialise the scratchpad */
	tmp.cf_conf = (char **) 0;
	tmp.cf_lconf = 0;
	tmp.cf_nconf = 0;

	/* fix up the configs for each mode */
	if (tcc_modes & TCC_BUILD)
		docfr2(&MCFLIST(CONF_BUILD), &tmp, sp, CONF_BUILD, "build");
	if (tcc_modes & TCC_EXEC)
		docfr2(&MCFLIST(CONF_EXEC), &tmp, sp, CONF_EXEC, "exec");
	if (tcc_modes & TCC_CLEAN)
		docfr2(&MCFLIST(CONF_CLEAN), &tmp, sp, CONF_CLEAN, "clean");

	/* free the scratchpad area */
	if (tmp.cf_conf) {
		TRACE2(tet_Tbuf, 6, "free tmp config list = %s",
			tet_i2x(tmp.cf_conf));
		free((char *) tmp.cf_conf);
	}
}

/*
**	docfr2() - extend the docfrem() processing for a particular mode
*/

static void docfr2(from, tp, sp, mode, cftype)
struct cflist *from;
register struct cflist *tp;
struct systab *sp;
int mode;
char *cftype;
{
	char fname[MAXPATH];
	char *texec, *tsroot;
	register char **cp;
	struct cflist *to;
	static char fmt[] = "tet_tcxconfig() failed when performing %s mode configuration variable exchange with system %d: rc =";
	char msg[sizeof fmt + LNUMSZ + 6];

	TRACE3(tet_Ttcc, 3,
		"docfr2(): fix up the %s configuration on system %s",
		cftype, tet_i2a(sp->sy_sysid));

	/* determine the config file name on the remote system */
	if ((tsroot = getdcfg("TET_TSROOT", sp->sy_sysid)) == (char *) 0) {
		error(0, "TET_TSROOT not defined for system",
			tet_i2a(sp->sy_sysid));
		conferrors++;
	}
	ASSERT(CONF_MODE_OK(mode, defcfname));
	if (mode == CONF_EXEC &&
		(texec = getdcfg("TET_EXECUTE", sp->sy_sysid)) != (char *) 0) {
			fullpath(texec, DEFCFNAME(mode), fname,
				sizeof fname, 1);
			if (tet_tcaccess(sp->sy_sysid, fname, 0) < 0)
				fullpath(tsroot, DEFCFNAME(mode), fname,
					sizeof fname, 1);
	}
	else
		fullpath(tsroot, DEFCFNAME(mode), fname, sizeof fname, 1);

	/*
	** build a list of configuration variables to send to the remote
	** system  which consists of all the generic variables plus all
	** the TET_REMnnn variables for that system
	*/
	tp->cf_nconf = 0;
	for (cp = from->cf_conf; cp < from->cf_conf + from->cf_nconf; cp++) {
		if (tet_remvar(*cp, -1) != *cp &&
			tet_remvar(*cp, sp->sy_sysid) == *cp)
				continue;
		RBUFCHK((char **) &tp->cf_conf, &tp->cf_lconf,
			(int) ((tp->cf_nconf + 1) * sizeof *tp->cf_conf));
		*(tp->cf_conf + tp->cf_nconf++) = *cp;
	}

	/* perform a config variable exchange with the remote system */
	ASSERT(CONF_MODE_OK(mode, sp->sy_cflist));
	to = &CFLIST(sp, mode);
	if (tet_tcxconfig(sp->sy_sysid, fname, tp, &vopts, to) < 0) {
		(void) sprintf(msg, fmt, cftype, sp->sy_sysid);
		error(0, msg, tet_ptrepcode(tet_tcerrno));
		conferrors++;
	}

	/* finally, check all of the boolean variables */
	checkbvar(to, sp->sy_sysid, cftype);
}

#endif /* !TET_LITE */	/* -END-LITE-CUT- */


/*
**	checkbvar() - check that all the boolean variables are
**		either "True" or "False" in a particular configuration
*/

static void checkbvar(lp, sysid, cftype)
struct cflist *lp;
int sysid;
char *cftype;
{
	/* list of boolean configuration variables */
	static struct bvar {
		char *bv_name;
		int bv_len;
	} bvar[] = {
		{ "TET_EXEC_IN_PLACE" },
		{ "TET_OUTPUT_CAPTURE" },
		{ "TET_API_COMPLIANT" },
		{ "TET_PASS_TC_NAME" },
		{ "TET_TRANSFER_SAVE_FILES" }
	};

#define Nbvar	(sizeof bvar / sizeof bvar[0])

	static char fmt[] = "bad value for boolean variable %s in %s configuration on system";
	char msg[sizeof fmt + 40];
	register char *p1, *p2;
	register char **cp;
	register struct bvar *vp;

	TRACE3(tet_Ttcc, 2, "checkbvar(): check the boolean variables in the %s configuration for system %s",
		cftype, tet_i2a(sysid));

	/* initialise the name lengths first time through */
	if (!bvar[0].bv_len)
		for (vp = bvar; vp < bvar + Nbvar; vp++)
			vp->bv_len = strlen(vp->bv_name);

	/*
	** search the list for boolean variables and check that each one is
	** either "true" or "false" (actually we only check for 't' and 'f')
	*/
	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++) {
		if ((p1 = tet_remvar(*cp, -1)) == (char *) 0 ||
			(p2 = tet_equindex(p1)) == (char *) 0)
				continue;
		for (vp = bvar; vp < bvar + Nbvar; vp++) {
			if (vp->bv_len != p2 - p1 ||
				strncmp(p1, vp->bv_name, vp->bv_len))
					continue;
			switch (*++p2) {
			case 'T':
			case 'F':
			case 't':
			case 'f':
				break;
			default:
				(void) sprintf(msg, fmt, vp->bv_name, cftype);
				error(0, msg, tet_i2a(sysid));
				conferrors++;
				break;
			}
			break;
		}
	}
}

/*
**	reportcfg() - report a configuration to the journal
*/

static void reportcfg(lp, sysid, mode)
struct cflist *lp;
int sysid, mode;
{
	TRACE3(tet_Ttcc, 3,
		"reportcfg(): report %s mode configuration for system %s",
		prcfmode(mode), tet_i2a(sysid));

	ASSERT_LITE(sysid == 0);

	/* emit the config start message */
	if (sysid == 0) {
		ASSERT(CONF_MODE_OK(mode, cfname));
		jnl_mcfg_start(CFNAME(mode), mode);
	}
#ifndef TET_LITE	/* -START-LITE-CUT- */
	else
		jnl_scfg_start(sysid, mode);
#endif /* !TET_LITE */	/* -END-LITE-CUT- */

	/* emit the config lines */
#ifndef TET_LITE	/* -START-LITE-CUT- */
	if (mode == CONF_DIST)
		rptdist(lp);
	else
#endif /* !TET_LITE */	/* -END-LITE-CUT- */
		rptconf(lp);

	/* emit the config end message */
	jnl_cfg_end();
}


#ifndef TET_LITE	/* -START-LITE-CUT- */

/*
**	rptdist() - reportcfg() processing for the distributed config
*/

static void rptdist(lp)
register struct cflist *lp;
{
	register int sysid, sysmax;
	register char **cp;
	register char *p;

	TRACE1(tet_Ttcc, 4, "rptdist(): report the distributed configuration");

	/* first, report all the TET_REMnnn_ variables in sysid order */
	for (sysid = 0, sysmax = symax(); sysid <= sysmax; sysid++)
		for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
			if ((p = tet_remvar(*cp, sysid)) != *cp) {
				ASSERT(p);
				jnl_cfg(*cp);
			}

	/* then report the others */
	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
		if (tet_remvar(*cp, -1) == *cp)
			jnl_cfg(*cp);
}

#endif /* !TET_LITE */	/* -END-LITE-CUT- */


/*
**	rptconf() - reportcfg() processing for a (non-DIST) config
*/

static void rptconf(lp)
register struct cflist *lp;
{
	register char **cp;

	TRACE1(tet_Ttcc, 4,
		"rptconf(): report a non-distributed configuration");

	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
		jnl_cfg(*cp);
}


/*
**	docffile() - write out a configuration to a temporary file
**		and return a pointer to the file name which is stored
**		in malloc'd memory
**
**	return (char *) 0 on error
*/

static char *docffile(lp, type)
struct cflist *lp;
char *type;
{
	char *fname;

	/* get a temporary file name */
	if ((fname = tet_mktfname("tcc")) == (char *) 0)
		return(fname);

	/* write out the variables to the file */
	if (docff2(fname, lp, type) < 0) {
		TRACE2(tet_Tbuf, 6, "free tmp config file name = %s",
			tet_i2x(fname));
		free(fname);
		fname = (char *) 0;
	}

	return(fname);
}

/*
**	docff2() - extend the docffile() processing
**
**	return 0 if successful or -1 on error
*/

static int docff2(fname, lp, type)
register struct cflist *lp;
char *fname, *type;
{
	register char **cp;
	FILE *fp;

	/* open the file */
	if ((fp = fopen(fname, "w")) == (FILE *) 0) {
		error(errno, "can't open", fname);
		return(-1);
	}

	/* write out the variables */
	(void) fprintf(fp, "# %s configuration variables\n\n", type);
	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
		if (fprintf(fp, "%s\n", *cp) < 0) {
			error(errno, "write error on", fname);
			(void) fclose(fp);
			return(-1);
		}

	/* close the file */
	if (fclose(fp) < 0) {
		error(errno, "close error on", fname);
		return(-1);
	}

	/* all OK so return success */
	return(0);
}

/*
**	getmcfg() - return pointer to value of config variable "name"
**		in the master configuration for the specified mode
**
**	return (char *) 0 if no entry appears for name
*/

char *getmcfg(name, mode)
char *name;
int mode;
{
	register char **cp, *p;

	mode = tcc2cfmode(mode);
	ASSERT(CONF_MODE_OK(mode, mcflist));

	if ((cp = findcfg(name, &MCFLIST(mode))) == (char **) 0)
		p = (char *) 0;
	else {
		p = tet_equindex(*cp);
		ASSERT(p);
		p++;
	}

	TRACE4(tet_Ttcc, 10, "getmcfg(\"%s\", %s) returns %s",
		name, prcfmode(mode), p ? p : "NULL");

	return(p);
}

/*
**	getdcfg() - return pointer to value of config variable "name"
**		in the distributed configuration for the specified system
*/

char *getdcfg(name, sysid)
char *name;
int sysid;
{
	register char **cp, *p1, *p2;
	register int len;
	register struct cflist *lp = &MCFLIST(CONF_DIST);

	ASSERT_LITE(sysid == 0);

	len = strlen(name);

	/* look for the variable with a TET_REMnnn_ prefix */
	for (cp = lp->cf_conf; cp < lp->cf_conf + lp->cf_nconf; cp++)
		if ((p1 = tet_remvar(*cp, sysid)) != (char *) 0 && p1 != *cp &&
			(p2 = tet_equindex(p1)) != (char *) 0 &&
			len == p2 - p1 && !strncmp(name, p1, len)) {
				++p2;
				TRACE4(tet_Ttcc, 10,
					"getdcfg(\"%s\", %s) returns %s", name,
					tet_i2a(sysid), p2 ? p2 : "NULL");
				return(p2);
		}

	/*
	** not found so look for the unadorned variable if sysid
	** refers to a remote system;
	** otherwise just return NULL if sysid refers to the local system
	**
	** this is because all the distributed config variables for the
	** local system are derived from environment variables;
	** we don't want a non-TET_REMnnn variable in the distributed
	** configuration supplying a default value when the corresponding
	** environment variable is undefined
	*/
	TRACE3(tet_Ttcc, 10, "getdcfg(\"%s\", %s): variable not found",
		name, tet_i2a(sysid));
	return(sysid > 0 ? getmcfg(name, CONF_DIST) : (char *) 0);
}

/*
**	getcfg() - return pointer to value of config variable "name"
**		in the per-system configuration for the specified mode
**
**	return (char *) 0 if no entry appears for name
*/

char *getcfg(name, sysid, mode)
char *name;
int sysid, mode;
{
	register char **cp, *p;
	register struct systab *sp;

	ASSERT_LITE(sysid == 0);

	/* get the systab entry for this system */
	sp = syfind(sysid);

	/*
	** if there is no systab entry for this system:
	**	if this is the local system, return the master config value;
	**	otherwise bail out on a programming error
	*/
	if (sp == (struct systab *) 0 && sysid == 0)
		return(getmcfg(name, mode));
	ASSERT(sp);

	/* convert mode if necessary and check it */
	mode = tcc2cfmode(mode);
	ASSERT(CONF_MODE_OK(mode, sp->sy_cflist));

	/* look up the variable in the per-system configuration */
	if ((cp = findcfg(name, &CFLIST(sp, mode))) == (char **) 0)
		p = (char *) 0;
	else {
		p = tet_equindex(*cp);
		ASSERT(p);
		p++;
	}

	TRACE5(tet_Ttcc, 10, "getcfg(\"%s\", %s, %s) returns %s",
		name, tet_i2a(sysid), prcfmode(mode), p ? p : "NULL");

	return(p);
}

/*
**	putdcfg() - add/update a distributed configuration variable
**		for a particular system
*/

void putdcfg(name, sysid, value)
char *name, *value;
int sysid;
{
	static char fmt[] = "TET_REM%03d_%s=%.*s";
	char buf[MAXPATH + sizeof fmt + 40];

	ASSERT_LITE(sysid == 0);

	(void) sprintf(buf, fmt, sysid % 1000, name,
		sizeof buf - strlen(name) - 14, value);
	proccfl2(buf, &MCFLIST(CONF_DIST), 1);
}

/*
**	getmcflag() - get the value of a configuration variable from the
**		specified master configuration list whose value can be
**		True or False
**
**	return 1 if the variable is set and its value is True, 0 otherwise
**
**	it is assumed that all flags default to False if unset
**	(currently this is the case)
*/

int getmcflag(name, mode)
char *name;
int mode;
{
	register char *p;

	if ((p = getmcfg(name, mode)) == (char *) 0)
		return(0);
	
	return(cflag2bool(name, p));
}

/*
**	getcflag() - get the value of a configuration variable from the
**		per-system configuration for the specified mode
**		whose value can be True or False
**
**	return 1 if the variable is set and its value is True, 0 otherwise
**
**	it is assumed that all flags default to False if unset
**	(currently this is the case)
*/

int getcflag(name, sysid, mode)
char *name;
int sysid, mode;
{
	register char *p;

	ASSERT_LITE(sysid == 0);

	if ((p = getcfg(name, sysid, mode)) == (char *) 0)
		return(0);
	
	return(cflag2bool(name, p));
}

/*
**	tcc2cfmode() - convert a mode of operation to a config mode
**
**	returns mode if mode is not a valid tcc mode of operation
*/

int tcc2cfmode(mode)
int mode;
{
	switch (mode) {
	case TCC_BUILD:
		return(CONF_BUILD);
	case TCC_EXEC:
		return(CONF_EXEC);
	case TCC_CLEAN:
		return(CONF_CLEAN);
	case CONF_BUILD:
	case CONF_EXEC:
	case CONF_CLEAN:
	case CONF_DIST:
		return(mode);
	default:
		/* this "can't happen" */
		fatal(0, "unexpected mode", tet_i2o(mode));
		/* NOTREACHED */
	}
}

/*
**	prcfmode() - return printable representation of a configuration
**		mode
*/

char *prcfmode(mode)
int mode;
{
	static char text[] = "conf-mode ";
	static char msg[sizeof text + LNUMSZ];

	switch (mode) {
	case CONF_BUILD:
		return("BUILD");
	case CONF_EXEC:
		return("EXEC");
	case CONF_CLEAN:
		return("CLEAN");
	case CONF_DIST:
		return("DIST");
	default:
		(void) sprintf(msg, "%s%d", text, mode);
		return(msg);
	}
}

/*
**	cflag2bool() - return the boolean value of a string which should
**		either be "True" or "False"
*/

static int cflag2bool(name, val)
char *name, *val;
{
	switch (*val) {
	case 'T':
	case 't':
		return(1);
	case 'F':
	case 'f':
		break;
	default:
		error(0, name, "variable has ambiguous value - False assumed");
		break;
	}

	return(0);
}

#ifdef TET_LITE	/* -LITE-CUT-LINE- */

/*
**	tet_config_putenv() - set up the value of the TET_CONFIG environment
**		variable which corresponds to the current mode of operation
**
**	return 0 if successful or -1 on error
*/

int tet_config_putenv(opmode)
int opmode;
{
	static int currmode = -1;
	static char *var;
	static int lvar;
	static char envname[] = "TET_CONFIG";
	int cfmode;

	cfmode = tcc2cfmode(opmode);
	ASSERT(CONF_MODE_OK(cfmode, cfname));

	/* return now if the current value has already been set */
	if (cfmode == currmode)
		return(0);

	/*
	** here to change the setting of TET_CONFIG in the environment
	**
	** ensure that the environment string is big enough,
	** then format the variable
	*/
	ASSERT(TCFNAME(cfmode));
	if (BUFCHK(&var, &lvar, (int) (sizeof envname + strlen(TCFNAME(cfmode)) + 1)) < 0)
		return(-1);
	(void) sprintf(var, "%s=%s", envname, TCFNAME(cfmode));

	/* put the variable in the environment */
	if (tet_putenv(var) < 0)
		return(-1);

	/* all ok so remember the current mode and return success */
	currmode = cfmode;
	return(0);
}

#endif /* TET_LITE */	/* -LITE-CUT-LINE- */



/*
**	confgiveup() - exit after finding configuration errors
*/

static void confgiveup()
{
	(void) fprintf(stderr,
		"%s: giving up after %d configuration error%s\n",
		tet_progname, conferrors, conferrors == 1 ? "" : "s");
	tcc_exit(1);
}

/*
**	config_cleanup() - remove temporary config files before exit
*/

void config_cleanup()
{

#ifdef TET_LITE		/* -LITE-CUT-LINE- */

	char **fname;

	for (fname = tcfname; fname < &tcfname[Ntcfname]; fname++)
		if (*fname)
			(void) UNLINK(*fname);

#else 	/* -START-LITE-CUT- */

	if (ecfname)
		(void) UNLINK(ecfname);
	if (dcfname)
		(void) UNLINK(dcfname);
	if (ccfname)
		(void) UNLINK(ccfname);

#endif /* TET_LITE */	/* -END-LITE-CUT- */

}

/*
**	get_runtime_tsroot() - return the name of the test suite root
**		directory to be used at runtime on a particular system
**
**	this function should not be called before distcfg() has set
**	up the distributed configurations for each system
**
**	the value returned by this function may not be useful before any
**	required runtime copy operations have been performed
**	(see the calls to rtlcopy() and rtrcopy() in tcc.c)
*/

char *get_runtime_tsroot(sysid)
int sysid;
{
	char *retval, *rtdir, *tsroot;
	static char rtsroot[MAXPATH];

	/*
	** if there is no TET_RUN, just return the system's TET_TSROOT;
	** otherwise we must construct the name of the tsroot directory
	** below TET_RUN and return that
	*/
	tsroot = getdcfg("TET_TSROOT", sysid);
	ASSERT(tsroot);
	if ((rtdir = getdcfg("TET_RUN", sysid)) == (char *) 0)
		retval = tsroot;
	else {
		fullpath(rtdir, tcc_basename(tsroot), rtsroot, sizeof rtsroot,
			sysid == 0 ? 0 : 1);
		retval = rtsroot;
	}

	TRACE3(tet_Ttcc, 10, "get_runtime_tsroot(%s) returns %s",
		tet_i2a(sysid), retval);
	return(retval);
}

