/*
 *	SCCS: @(#)getline.c	1.2 (96/11/04)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)getline.c	1.2 (96/11/04) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)getline.c	1.2 96/11/04 TETware release 3.2
NAME:		getline.c
PRODUCT:	TETware
AUTHOR:		Andrew Dingwall, UniSoft Ltd.
DATE CREATED:	August 1996

DESCRIPTION:
	functions to read a scenario file

MODIFICATIONS:

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "dtmac.h"
#include "error.h"
#include "bstring.h"
#include "llist.h"
#include "tcc.h"

#ifndef NOTRACE
#include "ltoa.h"
#endif


#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif

/* structure of the line cache stack used by getline() and ungetline() -
	the lc_next and lc_last elements must be 1st and 2nd so as to
	enable the stack to be manipulated by the llist functions */
struct lc {
	struct lc *lc_next;	/* ptr to next element in the list */
	struct lc *lc_last;	/* ptr to previous element in the list */
	char *lc_line;		/* the stored input line */
	int lc_lineno;		/* the input line number */
};

/* the line cache stack itself */
static struct lc *lcstack;

/* the stdio stream associated with the stack */
static FILE *sfp;

/* static function declarations */
#ifndef NOTRACE
static char *firstpart PROTOLIST((char *));
#endif
static struct lc *lcalloc PROTOLIST((void));
static void lcfree PROTOLIST((struct lc *));
static struct lc *lcpop PROTOLIST((void));
static void lcpush PROTOLIST((struct lc *));


/*
**	getline() - get the next non-blank, non-comment line of input
**
**	return a pointer to the line, or (char *) 0 on EOF or error
**
**	note that if ungetline() is to be used, only one file may be
**	active at one time!
*/

char *getline(fp, fname, lcp)
FILE *fp;
char *fname;
int *lcp;
{
	static char buf[LBUFLEN];
	struct lc *cp;
	register char *p;

	/* pop a line off the stack if there is one */
	if ((cp = lcpop()) != (struct lc *) 0) {
		ASSERT(fp == sfp);
		(void) strcpy(buf, cp->lc_line);
		*lcp = cp->lc_lineno;
		lcfree(cp);
		TRACE4(tet_Tscen, 10,
			"getline(): fp = %s, lineno = %s, line = <%s>",
			tet_i2x(fp), tet_i2a(*lcp), firstpart(buf));
		return(buf);
	}

	for (;;) {
		/* read the next line from the input stream */
		if (feof(fp) || fgets(buf, sizeof buf, fp) == (char *) 0) {
			if (ferror(fp))
				error(errno, "read error on", fname);
			ASSERT(lcpop() == (struct lc *) 0);
			sfp = (FILE *) 0;
			TRACE2(tet_Tscen, 10, "getline(): fp = %s, return EOF",
				tet_i2x(fp));
			return((char *) 0);
		}

		/* increment the line counter */
		++*lcp;

		/* strip comments (only in column 1) */
		if (buf[0] == '#')
			continue;

		/* strip the trailing newline */
		for (p = buf; *p; p++)
			if (*p == '\n') {
				*p = '\0';
				break;
			}

		/* strip trailing spaces */
		while (--p >= buf)
			if (isspace(*p))
				*p = '\0';
			else
				break;

		/* if there is anything left, return it */
		if (p >= buf) {
			TRACE4(tet_Tscen, 10,
				"getline(): fp = %s, lineno = %s, line = <%s>",
				tet_i2x(fp), tet_i2a(*lcp), firstpart(buf));
			return(buf);
		}
	}
}

/*
**	ungetline() - store a line for subsequent retrieval by getline()
*/

void ungetline(fp, line, lineno)
FILE *fp;
char *line;
int lineno;
{
	struct lc *cp;

	TRACE4(tet_Tscen, 10, "ungetline(): fp = %s, lineno = %s, line = <%s>",
		tet_i2x(fp), tet_i2a(lineno), firstpart(line));

	/* if this is a new file, ensure that the stack is empty */
	if (fp != sfp) {
		ASSERT(lcpop() == (struct lc *) 0);
		sfp = fp;
	}

	/* store the line and push it on to the stack */
	cp = lcalloc();
	cp->lc_line = rstrstore(line);
	cp->lc_lineno = lineno;
	lcpush(cp);
}

/*
**	lcalloc(), lcfree() - functions to allocate and free a
**		line cache element
*/

static struct lc *lcalloc()
{
	register struct lc *cp;

	errno = 0;
	if ((cp = (struct lc *) malloc(sizeof *cp)) == (struct lc *) 0)
		fatal(errno, "can't allocate line cache element", (char *) 0);

	TRACE2(tet_Tbuf, 6, "allocate lcache element = %s", tet_i2x(cp));
	bzero((char *) cp, sizeof *cp);
	return(cp);
}

static void lcfree(cp)
struct lc *cp;
{
	TRACE2(tet_Tbuf, 6, "free lcache element = %s", tet_i2x(cp));

	if (cp) {
		if (cp->lc_line) {
			TRACE2(tet_Tbuf, 6, "free lcache line = %s",
				tet_i2x(cp->lc_line));
			free(cp->lc_line);
		}
		free((char *) cp);
	}
}

/*
**	lcpush(), lcpop() - line cache stack manipulation functions
**
**	the operation of these functions depends on the LIFO behaviour
**	of linked lists which are manipulated by tet_listinsert() and
**	tet_listremove()
*/

static void lcpush(cp)
struct lc *cp;
{
	TRACE3(tet_Tscen, 10, "lcpush(): push line %s = <%s>",
		tet_i2a(cp->lc_lineno), cp->lc_line);

	tet_listinsert((struct llist **) &lcstack, (struct llist *) cp);
}

static struct lc *lcpop()
{
	register struct lc *cp;

	if ((cp = lcstack) != (struct lc *) 0) {
		tet_listremove((struct llist **) &lcstack, (struct llist *) cp);
		TRACE3(tet_Tscen, 10, "lcpop(): pop line %s from stack = <%s>",
			tet_i2a(cp->lc_lineno), cp->lc_line);
	}
	else
		TRACE1(tet_Tscen, 10, "lcpop(): stack is empty");

	return(cp);
}

/*
**	firstpart() - deliver the first part of a string
**
**	if the string is longer than PLEN, it has an eipsis appended
**	to it
*/

#ifndef NOTRACE

#define PLEN	24
static char *firstpart(s)
char *s;
{
	static char dots[] = " ...";
	static char buf[PLEN + sizeof dots];

	(void) sprintf(buf, "%.*s%s", PLEN, s,
		(int) strlen(s) > PLEN ? dots : "");

	return(buf);
}

#endif /* NOTRACE */

