/*
 *      SCCS:  @(#)child.c	1.21 (97/07/21) 
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1992 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 *
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)child.c	1.21 (97/07/21) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)child.c	1.21 97/07/21 TETware release 3.2
NAME:		child.c
PRODUCT:	TETware
AUTHOR:		Andrew Dingwall, UniSoft Ltd.
DATE CREATED:	October 1992

DESCRIPTION:
	tet_tcmc_main() is called from a different main() wrapper for each API,
	providing the main() function for processes started by tet_exec()
	and tet_remexec().

MODIFICATIONS:
	Andrew Dingwall, UniSoft Ltd., December 1993
	removed reference to dapi.h

	Geoff Clare, UniSoft Ltd., August 1996
	Changes for TETWare.

	Geoff Clare, UniSoft Ltd., 21 August 1996
	Made include of <stdio.h> non-conditional.

	Geoff Clare, UniSoft Ltd, Sept 1996
	Moved tet_blockable_sigs setup to tet_init_blockable_sigs().
	Changes for TETware-Lite.

	Geoff Clare, UniSoft Ltd., Oct 1996
	restructured tcm source to avoid "ld -r"

	Geoff Clare, UniSoft Ltd., July 1997
	Changes to support NT threads.

	Andrew Dingwall, UniSoft Ltd., July 1997
	Added code to generate a new (unique) context value on Win32 systems,
	since it is not possible for tet_spawn() to call tet_setcontext()
	on a Win32 system.

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "dtmac.h"
#include "dtmsg.h"
#include "dtthr.h"
#include "alarm.h"
#include "error.h"
#include "synreq.h"
#include "dtetlib.h"
#include "tcmfuncs.h"
#include "tet_api.h"
#include "apilib.h"


#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif


/* global variables, values inhereted from tet_exec() and tet_remexec() */
#ifdef TET_THREADS
extern thread_t tet_start_tid;
extern	void	tet_cln_threads();
extern	void	tet_mtx_init();
extern	void	tet_mtx_destroy();
#endif

#ifdef TET_THREADS
static	int	start_errno;
static	int	start_child;
static	int	start_alrm_flag;
static	long	start_block;
static	long	start_sequence;
#endif /* TET_THREADS */


int tet_tcmc_main(argc, argv)
int argc;
char **argv;
{
	register char **ap1, **ap2;
	register int rc;

#ifdef TET_THREADS
	/* initial thread ID */
	tet_start_tid = thr_self();

	/* set up thread-specific data */
	if (thr_keycreate(&tet_errno_key, TET_NULLFP) != 0 ||
	    thr_keycreate(&tet_child_key, TET_NULLFP) != 0 ||
	    thr_keycreate(&tet_alrm_flag_key, TET_NULLFP) != 0 ||
	    thr_keycreate(&tet_block_key, TET_NULLFP) != 0 ||
	    thr_keycreate(&tet_sequence_key, TET_NULLFP) != 0)
	{
		/* can't use tet_error() yet */
		(void) fprintf(stderr,
			"%s: thr_keycreate() failed in TCM startup",
			argc >= 5 ? argv[4] : argv[0]);
		exit(EXIT_FAILURE);
	}
	(void) thr_setspecific(tet_errno_key, (void *)&start_errno);
	(void) thr_setspecific(tet_child_key, (void *)&start_child);
	(void) thr_setspecific(tet_alrm_flag_key, (void *)&start_alrm_flag);
	(void) thr_setspecific(tet_block_key, (void *)&start_block);
	(void) thr_setspecific(tet_sequence_key, (void *)&start_sequence);

	/* initialise all mutexes */
	tet_mtx_init();
#endif

	if (argc < 5)
		fatal(0, "argument count wrong; process must be executed by",
			tet_callfuncname());

	/*
	** set the global variables from the command line -
	** on WIN32 systems there is no fork() in the parent process so
	** tet_setcontext() and tet_setblock() have not yet been called;
	** so argv[2] and argv[3] are the context and block values from the
	** parent and we must simulate calls to tet_setcontext() and
	** tet_setblock() in this process
	**
	** the pids get reused too quickly on WIN32 systems to be useful
	** as a context value;
	** we do the best we can to generate a unique context value
	** which can be used to distinguish journal output from that
	** of other processes that write to the same journal
	*/
	tet_thistest = atoi(argv[0]);
	tet_activity = atol(argv[1]);
	tet_context = atol(argv[2]);
#  ifdef TET_THREADS
	tet_next_block = atol(argv[3]);
#  else
	tet_block = atol(argv[3]);
#  endif

	/* shuffle the other arguments down */
	argc -= 4;
	ap1 = argv + 4;
	ap2 = argv;
	do {
		*ap2++ = *ap1++;
	} while (*ap1);
	*ap2 = NULL;
	tet_pname = argv[0];

	tet_init_blockable_sigs();

	/* initialise the server and transport stuff */
	tet_tcminit(argc, argv);

	/* start a new block and read in the config variables */
	tet_setblock();
	tet_config();

	/* call the user-supplied program */
	rc = tet_main(argc, argv);

#ifdef TET_THREADS
	tet_cln_threads(0);
	tet_mtx_destroy();
	tet_mtx_init();
#endif

	/* log off all the servers and exit */
	tet_logoff();
	return(rc);
}

/*
**	tet_ti_tcminit() - transport-independent initialisation for child TCM
*/

void tet_ti_tcminit()
{
	/* nothing */
}

