/*
 *	SCCS: @(#)ictp.c	1.20 (97/07/21)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

/*
 * Copyright 1990 Open Software Foundation (OSF)
 * Copyright 1990 Unix International (UI)
 * Copyright 1990 X/Open Company Limited (X/Open)
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of OSF, UI or X/Open not be used in 
 * advertising or publicity pertaining to distribution of the software 
 * without specific, written prior permission.  OSF, UI and X/Open make 
 * no representations about the suitability of this software for any purpose.  
 * It is provided "as is" without express or implied warranty.
 *
 * OSF, UI and X/Open DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO 
 * EVENT SHALL OSF, UI or X/Open BE LIABLE FOR ANY SPECIAL, INDIRECT OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF 
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR 
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef lint
static char sccsid[] = "@(#)ictp.c	1.20 (97/07/21) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)ictp.c	1.20 97/07/21 TETware release 3.2
NAME:		ictp.c
PRODUCT:	TETware
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	25 July 1990
DESCRIPTION:

	tet_icstart(), tet_icend(), tet_tpstart() and
	tet_tpend() are used by the TCM to output TCM start lines, IC
	start and end lines and test purpose start and result lines to
	the combined execution results file maintained by XRESD

MODIFICATIONS:
	
	June 1992
	DTET development - apilib/dresfile.c is derived from
	TET release 1.10

	Andrew Dingwall, UniSoft Ltd,. October 1992
	Some non-API functions specific to parent TCMs moved from
	apilib/dresfile.c to here.

	All vestages of the local execution results file and temporary
	execution results file removed - in the DTET, all the processing
	associated with these files is done by XRESD.

	All these functions completely re-written - auto-syncs are used
	to pass deletion and Abort information between TCMs as follows:

		if a TP is deleted within a TCM, that TCM votes NO at the
		TP start sync point
		if a TCM receives a NO sync vote at TP start, it knows that
		the TP has been deleted in another TCM

		if a result code action is Abort, XRESD returns ER_ABORT
		when the MTCM calls tet_xdtpend() from tet_tpend();
		the MTCM syncs with a NO vote to the end of the last TP

		STCMs will receive this NO vote at the start of any remaining
		TPs in the current IC, thus causing them to believe that
		these TPs have been deleted in the MTCM

		when STCMs receive the NO vote at the start of the next IC,
		they will interpret it as the signal to perform Abort
		processing

	Andrew Dingwall, UniSoft Ltd., November 1992
	Increased tpend sync timeout to 10 mins so as to allow for
	larger inbalance between test part execution times.

	Denis McConalogue, UniSoft Limited, June 1993
	API enhancements - moved all functions from [ms]tcmdist.c
			   to here. Added tet_ismaster().

	Denis McConalogue, UniSoft Limited, September 1993
	do not open tet_combined in tet_opencom() if an tet_xrid is
	already initialised.

	Andrew Dingwall, UniSoft Ltd., November 1994
	signal IC end from MTCM when result code action is Abort

	Geoff Clare, UniSoft Ltd., August 1996
	Changes for TETWare.

	Andrew Dingwall, UniSoft Ltd., August 1996
	changes for use with tetware tcc

	Geoff Clare, UniSoft Ltd., Sept 1996
	Changes for TETWare-Lite.

	Andrew Dingwall, UniSoft Ltd., June 1997
	changes to support the defined test case interface

************************************************************************/

#ifndef TET_LITE /* -START-LITE-CUT- */

#include <stdio.h>
#include <stdlib.h>
#  include <sys/types.h>
#  include <unistd.h>
#include <limits.h>
#include <errno.h>
#include "dtmac.h"
#include "error.h"
#include "dtmsg.h"
#include "dtthr.h"
#include "synreq.h"
#include "servlib.h"
#include "dtetlib.h"
#include "tcmfuncs.h"
#include "tet_api.h"
#include "tet_jrnl.h"
#include "apilib.h"

extern struct synreq *tet_synreq;	/* to receive sync reply info */

/* static function declarations */
static void tcm_xdicend PROTOLIST((int, int));
static void tcm_xdicstart PROTOLIST((int, int));
static void tcm_xdtpend PROTOLIST((int, int));
static void tcm_xdtpstart PROTOLIST((int, int));


/*
**	tet_icstart() - signal IC start
**
**	if the previous result code action was Abort, this function is only
**	executed by STCMs
*/

void tet_icstart(icno, tpcount)
int icno, tpcount;
{
	register struct synreq *sp;
	int nsys = tet_Nsname;
	int errflag;
	char errmsg[128];

	/* the MTCM informs XRESD of IC start */
	tcm_xdicstart(icno, tpcount);

	/* then all TCMs sync on IC start -
		if the MTCM votes NO, this means that the previous result
		code action was Abort */
	if (tet_tcm_async(MK_ASPNO(icno, 0, S_ICSTART), SV_YES, SV_SYNC_TIMEOUT,
		tet_synreq, &nsys) < 0) {
			(void) sprintf(errmsg,
				"Auto Sync failed at start of IC %d", icno);
			tet_error(tet_sderrno, errmsg);
			tet_exit(EXIT_FAILURE);
	}

	if (tet_sderrno == ER_OK)
		return;

	/* here if the auto sync failed in an expected way -
		a NO vote from the MTCM means that the test is to be aborted */
	errflag = 0;
	for (sp = tet_synreq; sp < tet_synreq + nsys; sp++)
		switch (sp->sy_state) {
		case SS_SYNCYES:
			break;
		case SS_SYNCNO:
			if (sp->sy_sysid == 0)
				break;
			/* else fall through */
		default:
			(void) sprintf(errmsg,
		"Auto sync error at start of IC %d, sysid = %d, state = %s",
				icno, sp->sy_sysid, tet_systate(sp->sy_state));
			tet_error(tet_sderrno, errmsg);
			errflag = 1;
			break;
		}


	if (errflag)
		tet_exit(EXIT_FAILURE);

	/* here if MTCM voted NO - cleanup and exit */
	tet_docleanup();
	tet_exit(EXIT_FAILURE);
}

/*
**	tet_icend() - signal IC end
*/

void tet_icend(icno, tpcount)
int icno, tpcount;
{
	/* the MTCM informs XRESD of IC end */
	tcm_xdicend(icno, tpcount);
}

/*
**	tet_tpstart() - signal TP start
*/

void tet_tpstart(icno, tpno, testnum)
int icno, tpno, testnum;
{
	register struct synreq *sp;
	register int vote;
	int nsys = tet_Nsname;
	int errflag;
	char errmsg[128];

#ifdef TET_THREADS
	tet_next_block = 1;
#endif
	tet_block = 1;
	tet_sequence = 1;

	/* the MTCM informs XRESD of TP start */
	tcm_xdtpstart(icno, testnum);

	/* see if the TP has been deleted in this TCM and set our vote
		accordingly */
	vote = tet_reason(testnum) ? SV_NO : SV_YES;

	/* then all the TCMs sync on TP start */
	if (tet_tcm_async(MK_ASPNO(icno, tpno, S_TPSTART), vote,
		SV_SYNC_TIMEOUT, tet_synreq, &nsys) < 0) {
			(void) sprintf(errmsg,
				"Auto Sync failed at start of TP %d", testnum);
			tet_error(tet_sderrno, errmsg);
			tet_exit(EXIT_FAILURE);
	}

	if (tet_sderrno == ER_OK)
		return;

	/* here if sync failed in an expected way -
		a NO vote means that the TP has been deleted in another TCM */
	errflag = 0;
	for (sp = tet_synreq; sp < tet_synreq + nsys; sp++)
		switch (sp->sy_state) {
		case SS_SYNCYES:
			break;
		case SS_SYNCNO:
			tet_sync_del = 1;
			break;
		default:
			(void) sprintf(errmsg,
		"Auto Sync error at start of TP %d, sysid = %d, state = %s",
				testnum, sp->sy_sysid,
				tet_systate(sp->sy_state));
			tet_error(tet_sderrno, errmsg);
			errflag = 1;
			break;
		}

	if (errflag)
		tet_exit(EXIT_FAILURE);
}

void tet_tpend(icno, tpno, testnum)
int icno, tpno, testnum;
{
	register struct synreq *sp;
	int err;
	int nsys = tet_Nsname;
	char errmsg[128];

	/* all the TCMs sync YES on TP end -
		there is an assumption here that the parts of a TP will
		arrive at their ends within 10 minutes of each other;
		if this is not so for a particular test, the test writer
		is expected to use tet_sync() with a longer timeout
		to delay the ends of the quicker test parts */
	if (tet_tcm_async(MK_ASPNO(icno, tpno, S_TPEND), SV_YES,
		SV_SYNC_TIMEOUT * 10, tet_synreq, &nsys) < 0) {
			(void) sprintf(errmsg, 
				"Auto Sync failed at end of TP %d", testnum);
			tet_error(tet_sderrno, errmsg);
			tet_exit(EXIT_FAILURE);
	}
	err = tet_sderrno;

	/* then the MTCM informs XRESD of TP end -
		if XRESD returns ER_ABORT when tet_xdtpend() is called, this
		function does not return in the MTCM */
	tcm_xdtpend(icno, testnum);

	if (err == ER_OK)
		return;

	/* here if a TCM voted NO ("can't happen"), timed out or died */
	for (sp = tet_synreq; sp < tet_synreq + nsys; sp++)
		switch (sp->sy_state) {
		case SS_SYNCYES:
			break;
		default:
			(void) sprintf(errmsg,
		"Auto Sync error at end of TP %d, sysid = %d, state = %s",
				testnum, sp->sy_sysid,
				tet_systate(sp->sy_state));
			tet_error(err, errmsg);
			break;
		}

	tet_exit(EXIT_FAILURE);
}

/*
**	tcm_xdicstart() - inform XRESD of IC start from MTCM
*/

static void tcm_xdicstart(icno, tpcount)
int icno, tpcount;
{
	if (!tet_ismaster()) 
		return;

	if (tet_xdicstart(tet_xrid, icno, tet_activity, tpcount) < 0) {
		tet_error(tet_xderrno, "can't inform XRESD of IC start");
		tet_exit(EXIT_FAILURE);
	}
}

/*
**	tcm_xdicend() - inform XRESD of IC end from MTCM
*/

/* ARGSUSED */
static void tcm_xdicend(icno, tpcount)
int icno, tpcount;
{
	if (!tet_ismaster()) 
		return;

	if (tet_xdicend(tet_xrid) < 0) {
		tet_error(tet_xderrno, "can't inform XRESD of IC end");
		tet_exit(EXIT_FAILURE);
	}
}

/*
**	tcm_xdtpstart() - inform XRESD of TP start from MTCM
*/

/* ARGSUSED */
static void tcm_xdtpstart(icno, tpno)
int icno, tpno;
{
	if (!tet_ismaster()) 
		return;

	if (tet_xdtpstart(tet_xrid, tpno) < 0) {
		tet_error(tet_xderrno, "can't inform XRESD of TP start");
		tet_exit(EXIT_FAILURE);
	}
}

/*
**	tcm_xdtpend() - inform XRESD of TP end from MTCM
**
**	this function does not return in the MTCM if XRESD indicates
**	that the testcase must be aborted
*/

/* ARGSUSED */
static void tcm_xdtpend(icno, tpno)
int icno, tpno;
{
	char errmsg[128];
	struct synreq *synreq;
	int nsys, errflag;
	register struct synreq *sp;

	if (!tet_ismaster()) 
		return;

	/* signal TP end to XRESD */
	(void) tet_xdtpend(tet_xrid);
	switch (tet_xderrno) {
	case ER_OK:
		return;
	case ER_ABORT:
		break;
	default:
		tet_error(tet_xderrno, "can't inform XRESD of TP end");
		tet_exit(EXIT_FAILURE);
		/* NOTREACHED */
	}

	/* here if previous TP result code action was Abort -
		allocate memory for synreq array; don't worry if it fails */
	errno = 0;
	if ((synreq = (struct synreq *) malloc((size_t)(sizeof *synreq * tet_Nsname))) == (struct synreq *) 0) {
		tet_error(errno, " can't allocate synreq array for Abort sync");
		nsys = 0;
	}
	else
		nsys = tet_Nsname;

	/* signal IC end to XRESD */
	if (tet_xdicend(tet_xrid) < 0)
		tet_error(tet_xderrno, "ABORT: can't inform XRESD of IC end");

	/* sync NO to the end of the last IC -
		this communicates the Abort action to other TCMs */
	if (tet_tcm_async(MK_ASPNO(tet_iclast, ~0, S_ICEND), SV_NO,
		SV_SYNC_TIMEOUT, synreq, &nsys) < 0) {
			tet_error(tet_sderrno, "Abort Auto Sync failed");
			tet_exit(EXIT_FAILURE);
	}

	/* make sure that the other TCMs are still alive */
	if (synreq) {
		errflag = 0;
		for (sp = synreq; sp < synreq + nsys; sp++)
			switch (sp->sy_state) {
			case SS_SYNCYES:
			case SS_SYNCNO:
				break;
			default:
				(void) sprintf(errmsg,
			"Abort Auto Sync error, sysid = %d, state = %s",
					sp->sy_sysid,
					tet_systate(sp->sy_state));
				tet_error(tet_sderrno, errmsg);
				errflag = 1;
				break;
			}
		if (errflag)
			tet_exit(EXIT_FAILURE);
	}

	/* here if we should attempt to call (*tet_cleanup)() */
	tet_docleanup();
	tet_exit(EXIT_FAILURE);
}

/*
**	tet_tcmstart() - send TCM Start journal line to XRESD
*/

void tet_tcmstart(versn, no_ics)
char *versn;
int no_ics;
{
	char buf[128];

	if (!tet_ismaster()) 
		return;

	(void) sprintf(buf, "%d|%ld %s %d|TCM Start", 
		TET_JNL_TCM_START, tet_activity, versn, no_ics);

	if (tet_xdxres(tet_xrid, buf) < 0) {
		tet_error(tet_xderrno,
			"can't send \"TCM Start\" journal line to XRESD");
		tet_exit(EXIT_FAILURE);
	}
}

/*
**	tet_ismaster() - return true if we are running on the master
**			 system.
*/

int tet_ismaster()
{
	extern int tet_myptype;

	if (tet_myptype == PT_MTCM)
		return (1);
	else
		if (tet_myptype == PT_STCM)
			return (0);
		else {
			tet_error(0, "Can't determine system type");
			tet_exit(EXIT_FAILURE);
		}
		/* NOTREACHED */
}

#else /* -END-LITE-CUT- */

/* avoid "empty" file */
int tet_ictp_dot_c_not_needed;

#endif /* -LITE-CUT-LINE- */

